package listactors

import scala.actors._
import scala.actors.Actor._
import scala.actors.Future


  /**
   * Die Case Classes dienen hier der Nachrichtenbermittlung an den ListCalculations Actor
   * SimpleCalculation, AsyncCalculation und FutureCalculation berbringen jeweils
   * einen MapListProcessor um die Elemente einer Liste auf neue Werte zu mappen.
   * Ob dieses Mappen nebenlufig geschieht hngt vom Case ab (Fallunterscheidung siehe in receive)
   *   
   */

case class SimpleCalculation ( lp : MapListProcessor )
case class AsyncCalculation  ( lp: MapListProcessor , numberOfActors : Int )
case class FutureCalculation ( lp: MapListProcessor , numberOfActors : Int )
case class SetResultList (l : List [Int] , name: String)


  /**
   *  ListCalculations sind Objekte mit zwei Listen
   *    data: diese Liste enthlt die Ausgangswerte
   *    	  neue Werte werden ber Int - Nachrichten hinzugefgt, wobei
   *          der gesendete Int Wert durch eine modifier Funktion schon in einen
   *          anderern Startwert umgewandelt wird.
   *          Sendet man etwa die Werte 1 , 2 , 3 und nutzt einen Modifier, der jeden
   *          Wert mit 2 multipliziert (x => x *2), dann ergibt dies List ( 6 , 4, 2),
   *          da jeder neu hinzugefgte Wert mit 2 multipliziert und vor die Liste als neuer
   *          Kopf gesetzt wird.
   *          
   *    resultData: dieses Feld soll spter die gemappte Liste enthalten 
   *     
   *          
   */


class ListCalculations ( idName : String , modifier : Int => Int) extends Actor{

   start()

   
   private var data : List [Int] = List()			// wird mit Startwerten gefllt via Int Nachrichten
   private var resultData : List [Int] = List()     // wird mit setResult gesetzt, sobald das Ergebnis vorliegt


   // Implementierung des Actors:
   def act () {
     
     while (true) {
       

       // Empfangen von Int, SimpleCalculation, AsyncCalculation, 
       // FutureCalculation, SetResultList - Nachrichten 
       // TODO: Bitte implementieren Sie die Funktionen, mit denen auf die
       //       AsyncCalculation und FutureCalculation - Nachrichten reagiert wird
       
       receive {
         case value : Int => data = modifier (value) :: data 
     
         case sc: SimpleCalculation => setResult ( sc.lp.process(data) ,"SC-"+sc.lp.name)           // ACHTUNG - WAS IST HIER KRITISCH?!?	 
         
         // SOLUTION: Asynchrone Lsung
	     case ac: AsyncCalculation => {
	       
	       // Da wir die Gesamtliste in Teillisten aufteilen, und dann fr jede Teilliste die Berechnung parallel und asynchron durchfhren,
	       // mssen wir die Ergebnisse irgendwo sammeln. Hierzu erzeugen wir einen speziellen Actor, der die Teilergebnisse sammelt:
	       val resultCollector = new ResultCollector (ac.numberOfActors , ac.lp.name , this)
	       resultCollector.start()
	       
	       // Ermittel der Segmentgre
	       val segmentSize = if (ac.numberOfActors > 0 && ac.numberOfActors <= data.length)  data.length / ac.numberOfActors  else data.length
      
	       // Aufteilen der Gesamtliste in Teillisten durch Iteration (hier einmal nicht funktional :(
	       var tmpData = data  // Wird genutzt um, die Liste immer weiter zu verkrzen, also mit der Restliste arbeiten
	       var position = 0    // Sicherstellen, dass nie zu viele Listen erstellt werden (wegen Rundung) 
	       while (tmpData.length > segmentSize && position < ac.numberOfActors - 1) {
	          val split =  tmpData.splitAt(segmentSize)
	          // Erster Teil wird zum Berechnen geschickt:
	          startAsyncCalc ( split._1 , position ,  ac.lp , resultCollector )
	          
	          // zweiter Teil wird nun Rest der Liste:
	          tmpData = split._2
	          position += 1
	       }
	       // Restliste auch noch berechnen:
	       startAsyncCalc ( tmpData , ac.numberOfActors -1 ,  ac.lp , resultCollector )

	       
	       // Jetzt sind alle Teilberechnungen gestartet. Die Ausgabe erfolgt nicht hier, sondern es ist Aufgabe des ResultCollectors
	       // diesem Objekt (dem ListCalculations-Objekt) eine Nachricht mit dem fertigen Gesamtresultat zu senden.
	       
	       
	     }

	     
	     // SOLUTION: Lsung mit Futures
	     case fc: FutureCalculation => {
	       
	       	  // Auch hier bentigen wir die Segmentgre in Abhngigkeit der geforderten Actors
	          val segmentSize = if (fc.numberOfActors > 0 && fc.numberOfActors <= data.length)  data.length / fc.numberOfActors  else data.length
	          
	          // In der folgenden Variable speichern wir jeweils eine Liste mit allen Future-Objekten. 
	          // Da eine Futureobjekt etwas beliebiges als Resultat liefern kann nehmen wir hier eine Liste mit Future[Any]
	          var futures : List [Future[Any]] = List()

	          // Auch hier wird die Liste in Teillisten aufgeteilt. Eventuell wird ein Actor mehr als angegeben erzeugt (um es einfacher darzustellen).
	          var tmpData = data;
	          while (tmpData.length > 0 ) {
	            
	            // Nchstes Segment oder die gesamte Liste als Teilliste verwenden:
	            val splitPos = if (tmpData.length > segmentSize) segmentSize else tmpData.length
	            val splitted = tmpData.splitAt(splitPos)
	            tmpData = splitted._2

	            // Fr jede Teilliste starten wir einen CoWorker. Dem CoWorker geben wir die Teilliste und einen MapListProcessor
	            // Der CoWorker arbeitet los sobald er die Task-nachricht erhlt. Da wir noch nicht auf das Ergebnis zugreifen wird noch
	            // nicht blockiert.
	            val cw = new CoWorker()
	            val fut = cw !! Task ( splitted._1 , fc.lp)
	   
	            // Damit wir spter auf das Ergebnis zugreifen knnen, mssen wir uns alle Futures in einer Liste merken,
	            // dazu erzeugen wir eine neue Liste mit dem zustzlichen Future-Objekt
	            futures = futures:+fut
	            
	          }
	          
	          // An dieser Stelle haben wir alle Teil-Liste an CoWorker Actors geschickt. Die Berechnung luft. 
	          // Nun wollen wir die Werte einsammeln.
	          
	          // Fr jedes Future-Objekt fragen wir das Eregebnis mit future() ab. Der Aufruf blockiert solange bis das Ergebnis auch wirklich
	          // vorliegt. Da wir die futures in der Liste in der richtigen Reihenfolge gespeichert haben, mssen wir uns nicht mehr um
	          // die richtige Positionierung kmmern. Da wir das Ergebnis eines Futures nicht kennen (je nach nachricht knnten unterschiedliche
	          // Werte zurckgeschickt werden) mssen wir hier einen explizten Cast vornehmen. Wir wissen ja, dass der CoWorker nur Int-Listen zurckschickt:
	      
	          val preRes = for ( future <- futures) yield future().asInstanceOf [List[Int]] 
	          
	          // preRes enthlt nun eine Liste mit den Teilergebnissen (jeweils wieder Listen). Jetzt machen wir daraus eine flache Liste:
	          val res = preRes.flatten
	          
	       
	          setResult ( res , "FC-"+fc.lp.name )
	       
	     
	     }
	     case result:SetResultList => {  setResult ( result.l ,"ASC-"+ result.name)  }
         
         
       }
     }
    }   
   
   
  /**
   * Mit dieser Methode wird die Ergebnis-Liste gesetzt sobald diese vorliegt.
   * Als Nebeneffekt werden die ursprngliche und die neue Liste auf der Konsole ausgegeben,
   * damit Sie die Ergebnisse besser berprfen knnen.
   * @param result eine Liste mit neuberechneten (gemappten) Int Werten
   * @param method beschreibender Text - um nachzuvollziehen mit welcher Vorgehensweise das Ergebnis berechnet wurde, nur fr den Seiteneffekt
   */   
   
   private def setResult (result : List[Int] , method : String) {
      resultData = result      
      println ( method + " of " + this )
     
   }
  
   
   /***
    * Erhlt eine Teilliste und startet dafr den Actor
    */
   private def startAsyncCalc(partList : List [Int] , position : Integer, mlp:  MapListProcessor , collector : ResultCollector ) {
     
       val neuerActor = actor {
            val partResult = mlp.process(partList);
            collector ! PartResult (partResult , position)
            // println ("Result at pos " + position + "  is " + partResult);

       }
     
     
   }
   
   
   override def toString = idName + ": " + data.toString() + " --> " + resultData.toString()
 
   
   
   
}